# (c) Copyright 2019 Trent Hauck
# All Rights Reserved
"""The base tokenizer from which other tokenizers are expected to inhereit.

The `SequenceTokenizerSettings` in particular holds common tokens and their associated ids for
common special tokens. For example `unk_token` for unknown tokens.
"""

from typing import List
from typing import Optional

from pydantic import BaseSettings
from pydantic import Field
from pydantic import root_validator

from gcgc.vocab import Vocab


def _pad_token_list(tokens: List[str], pad_to: int, pad_char: str, pad_at_end: bool):
    """Pad the input tokens list to tokens."""
    token_len = len(tokens)

    if token_len > pad_to:
        return tokens

    new_tokens = pad_to - token_len

    padding = [pad_char] * new_tokens
    if pad_at_end:
        return tokens + padding
    return padding + tokens


class SequenceTokenizer(BaseSettings):
    """The sequence tokenizer object."""

    vocab: Vocab = Vocab()

    pad_token: Optional[str] = Field("|", env="GCGC_PAD_TOKEN")
    bos_token: Optional[str] = Field(">", env="GCGC_BOS_TOKEN")
    eos_token: Optional[str] = Field("<", env="GCGC_EOS_TOKEN")
    mask_token: Optional[str] = Field("#", env="GCGC_MASK_TOKEN")
    unk_token: Optional[str] = Field("?", env="GCGC_UNK_TOKEN")

    pad_token_id: Optional[int] = Field(None, env="GCGC_PAD_TOKEN_ID")
    bos_token_id: Optional[int] = Field(None, env="GCGC_BOS_TOKEN_ID")
    eos_token_id: Optional[int] = Field(None, env="GCGC_EOS_TOKEN_ID")
    mask_token_id: Optional[int] = Field(None, env="GCGC_MASK_TOKEN_ID")
    unk_token_id: Optional[int] = Field(None, env="GCGC_UNK_TOKEN_ID")

    pad_at_end: bool = Field(True, env="GCGC_PAD_AT_END")

    max_length: Optional[int] = Field(None, env="GCGC_MAX_LENGTH")
    min_length: Optional[int] = Field(None, env="GCGC_MIN_LENGTH")
    conform_length: Optional[int] = Field(None, env="GCGC_CONFORM_LENGTH")

    @classmethod
    def bare_tokenizer(cls, **values):
        """Init a tokenizer like normal, but default all tokens to None."""
        pad_token = values.pop("pad_token", None)
        bos_token = values.pop("bos_token", None)
        eos_token = values.pop("eos_token", None)
        mask_token = values.pop("mask_token", None)
        unk_token = values.pop("unk_token", None)

        return cls(
            pad_token=pad_token,
            bos_token=bos_token,
            eos_token=eos_token,
            mask_token=mask_token,
            unk_token=unk_token,
            **values,
        )

    @root_validator
    def token_ids(cls, values):  # pylint: disable=no-self-use,no-self-argument
        """Update the pad token id if appropriate."""
        vocab = values["vocab"]

        for key, value in values.items():
            if key.endswith("_token"):
                token_id_field = f"{key}_id"
                token_id = values[token_id_field]

                if token_id is None and value is not None:
                    vocab.add_item(value)
                    values[token_id_field] = vocab[value]
                elif token_id is not None and value is not None:
                    # pylint: disable=fixme
                    # TODO: check that token_id isn't already set (maybe its in the inverse?)
                    vocab[value] = token_id

        return values

    @root_validator
    def validate_lengths(cls, values):  # pylint: disable=no-self-argument, no-self-use
        """Check the length arguments are valid."""
        max_length = values.get("max_length")
        min_length = values.get("min_length")
        conform_length = values.get("conform_length")
        pad_token = values.get("pad_token")

        if conform_length is not None and (max_length is not None or min_length is not None):
            raise ValueError("If conform length is not None, max and min length can't be set.")

        if max_length is not None and min_length is not None and min_length > max_length:
            raise ValueError(
                f"Min length {min_length} cannot be more than max length {max_length}."
            )

        if (conform_length is not None or min_length is not None) and pad_token is None:
            raise ValueError("Cannot set conform_length or min_length without a pad_token.")

        return values

    @property
    def special_token_ids(self) -> List[int]:
        """Return the list of token ids corresponding to special tokens."""
        return [
            x
            for x in [
                self.bos_token_id,
                self.eos_token_id,
                self.unk_token_id,
                self.pad_token_id,
                self.mask_token_id,
            ]
            if x is not None
        ]

    @property
    def special_tokens(self) -> List[str]:
        """Return the set of special tokens that are not None."""
        return [
            x
            for x in [
                self.bos_token,
                self.eos_token,
                self.unk_token,
                self.pad_token,
                self.mask_token,
            ]
            if x is not None
        ]

    def get_special_tokens_mask(self, token_ids: List[int]) -> List[int]:
        """Given the input set of tokens, return a list that demarcates special character.

        ```python
        # Assuming 1 is bos_token, 2 is eos_token, and 0 is pad_token.
        >>> tokenizer.get_special_tokens_mask([1, 34, 21, 0, 0, 0, 2])
        [1, 0, 0, 1, 1, 1, 1]
        ```

        Args:
            token_ids: The list of integer tokens that may or may not be special tokens according
                to the toeknizer's settings.

        Returns:
            A list of 0s and 1s, where the value is one if the token is a special character.

        """
        return [int(token_id in self.special_token_ids) for token_id in token_ids]

    def apply_length_constraints(self, tokens: List[str]) -> List[str]:
        """Apply the constraints from the settings to the passed tokens list."""
        if self.conform_length:
            tokens = _pad_token_list(
                tokens, self.conform_length, str(self.pad_token), self.pad_at_end,
            )
            return tokens[: self.conform_length]

        if self.min_length:
            tokens = _pad_token_list(tokens, self.min_length, str(self.pad_token), self.pad_at_end)

        if self.max_length:
            tokens = tokens[: self.max_length]

        return tokens



# (c) Copyright 2019 Trent Hauck
# All Rights Reserved
"""This module holds the kmer tokenizer and its settings.

The `KmerTokenizer` tokenizers incoming sequences into kmers of configurable size against a
configurable alphabet.

For example, given incoming nucleotide sequences, using settings of kmer length 3 and stride 3 they
encoded sequences will be codons (in the loose sense) with a vocabulary of size 64.

```python

>>> tokenizer.encode('AAATTTCCC')
[0, 42, 63]

>>> len(tokenizer.vocab)
64
```
"""

import itertools as it
from typing import List

from pydantic import Field
from pydantic import root_validator
from pydantic import validator

from gcgc import alphabets
from gcgc.tokenizer.base import SequenceTokenizer
from gcgc.vocab import Vocab


def _create_kmer_vocab_from_token(vocab, alphabet: str, kmer_length: int) -> Vocab:
    """Create the vocab object from a list of tokens."""

    token_set = ["".join(kmer) for kmer in it.product(list(alphabet), repeat=kmer_length)]

    for token in token_set:
        vocab.add_item(token)

    return vocab


class KmerTokenizer(SequenceTokenizer):
    """The Kmer Tokenizer that encodes sequences into chunked kmers."""

    alphabet: str = Field("ATCG", env="GCGC_ALPHABET")
    kmer_length: int = Field(1, env="GCGC_KMER_LENGTH")
    kmer_stride: int = Field(1, env="GCGC_KMER_STRIDE")

    @root_validator
    def augment_vocab_with_kmer(cls, values):  # pylint: disable=no-self-argument,no-self-use
        """Update the vocab of the SequenceTokenizer with a kmer alphabet."""
        vocab = values["vocab"]
        alphabet = values["alphabet"]
        kmer_length = values["kmer_length"]

        values["vocab"] = _create_kmer_vocab_from_token(vocab, alphabet, kmer_length)
        return values

    # pylint: disable=no-self-use, no-self-argument
    @validator("alphabet")
    def resolve_alphabet(cls, alphabet: str) -> str:
        """Resolve the alphabet if it's a named alphabet.

        Args:
            alphabet: The raw alphabet, either the sequence literal or a name of a preset alphabet.

        Returns:
            The new alphabet.

        """
        return alphabets.resolve_alphabet(alphabet)

    def _kmer_n(self, seq: str) -> List[str]:
        seq_len = len(seq)
        iterations = seq_len - self.kmer_length + 1
        kmers = []

        for i in range(0, iterations, self.kmer_stride):
            kmer = seq[i : i + self.kmer_length]
            kmers.append(kmer)

        return kmers

    def encode(self, seq: str, add_unknown: bool = False) -> List[int]:
        """Encode the underlying sequence into a list of tokens ids.

        Args:
            seq: The incoming sequence to encode.
            add_unknown: If True, add the unknown token rather than throwing an out of vocabulary
                error.

        Returns:
            A list of the encoded tokens.

        """
        encoded = []

        for letter in self.tokenize(seq):
            try:
                encoded.append(self.vocab[letter])
            except KeyError:
                if add_unknown:
                    encoded.append(self.unk_token_id)
                else:
                    raise

        return encoded

    def tokenize(self, seq: str) -> List[str]:
        """Tokenize the sequence into a list of tokens.

        Args:
            seq: The sequence to encode.

        Returns:
            The list of strings that are the tokens.

        """
        seq_len = len(seq)

        if seq_len < self.kmer_length:
            raise ValueError(
                f"seq length {seq_len} cannot be less than the kmer size {self.kmer_length}"
            )

        if self.kmer_length == 1:
            kmer_list = list(seq)
        else:
            kmer_list = self._kmer_n(seq)

        if self.bos_token:
            kmer_list = [self.bos_token] + kmer_list

        if self.eos_token:
            kmer_list = kmer_list + [self.eos_token]

        return super().apply_length_constraints(kmer_list)



# (c) Copyright 2018 Trent Hauck
# All Rights Reserved
"""Top-level GCGC module."""

import warnings as _warnings

from gcgc import tokenizer
from gcgc.tokenizer.kmer_tokenzier import KmerTokenizer
from gcgc.tokenizer.kmer_tokenzier import SequenceTokenizer

__all__ = ["tokenizer", "KmerTokenizer", "SequenceTokenizer"]

__version__ = "0.12.2-dev.6"

_warnings.simplefilter(action="ignore", category=PendingDeprecationWarning)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['gcgc',
 'gcgc.data',
 'gcgc.tests',
 'gcgc.tests.fixtures',
 'gcgc.tests.tokenizer',
 'gcgc.tests.vocab',
 'gcgc.tokenizer']

package_data = \
{'': ['*'],
 'gcgc.data': ['splice/*'],
 'gcgc.tests.fixtures': ['PF12057/*',
                         'ecoli/*',
                         'globin_alignment/*',
                         'p53_human/*']}

install_requires = \
['pydantic>=1,<2']

extras_require = \
{'third_party': ['biopython>=1,<2', 'transformers>=2', 'torch>=1.5']}

setup_kwargs = {
    'name': 'gcgc',
    'version': '0.12.2.dev6',
    'description': 'GCGC is a preprocessing library for biological sequence model development.',
    'long_description': '# GCGC\n\n> GCGC is a tool for feature processing on Biological Sequences.\n\n[![](https://github.com/tshauck/gcgc/workflows/Run%20Tests%20and%20Lint/badge.svg)](https://github.com/tshauck/gcgc/actions?query=workflow%3A%22Run+Tests+and+Lint%22)\n[![](https://img.shields.io/pypi/v/gcgc.svg)](https://pypi.python.org/pypi/gcgc)\n[![DOI](https://zenodo.org/badge/DOI/10.5281/zenodo.2329966.svg)](https://doi.org/10.5281/zenodo.2329966)\n[![code style black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/psf/black)\n\n## Installation\n\nGCGC is primarily intended to be used as part of a larger workflow inside\nPython, but it can also be used as a docker container.\n\nTo install via pip:\n\n```sh\n$ pip install gcgc\n```\n\nAnd to pull the docker image:\n\n```sh\n$ docker pull docker.io/thauck/gcgc\n```\n\n## Documentation\n\nThe GCGC documentation is at [gcgc.trenthauck.com](http://gcgc.trenthauck.com),\nplease see it for an example.\n\n## Citing GCGC\n\nIf you use GCGC in your research, cite it with the following:\n\n```\n@misc{trent_hauck_2018_2329966,\n  author       = {Trent Hauck},\n  title        = {GCGC},\n  month        = dec,\n  year         = 2018,\n  doi          = {10.5281/zenodo.2329966},\n  url          = {https://doi.org/10.5281/zenodo.2329966}\n}\n```\n',
    'author': 'Trent Hauck',
    'author_email': 'trent@trenthauck.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'http://gcgc.trenthauck.com/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



# (c) Copyright 2020 Trent Hauck
# All Rights Reserved
"""Test the vocabulary object."""

from gcgc.vocab import Vocab


def test_vocab():
    """Test the vocabulary object works as expected."""
    vocab = Vocab()

    vocab.add_item("A")

    assert "A" in vocab
    assert vocab["A"] == 0

    vocab.add_items(["T", "C"])

    assert "T" in vocab
    assert "C" in vocab

    assert vocab.stoi == {"A": 0, "T": 1, "C": 2}
